@@ -1038,8 +1038,13 @@ var ReadStream = fs.ReadStream = function(path, options) {
   if (this.encoding) this.setEncoding(this.encoding);
 
   if (this.start !== undefined) {
+    if ('number' !== typeof this.start) {
+      throw TypeError('start must be a Number');
+    }
     if (this.end === undefined) {
       this.end = Infinity;
+    } else if ('number' !== typeof this.end) {
+      throw TypeError('end must be a Number');
     }
 
     if (this.start > this.end) {
@@ -1050,6 +1055,7 @@ var ReadStream = fs.ReadStream = function(path, options) {
   }
 
   if (this.fd !== null) {
+    this._read();
     return;
   }
 
@@ -1225,6 +1231,9 @@ var WriteStream = fs.WriteStream = function(path, options) {
   }
 
   if (this.start !== undefined) {
+    if ('number' !== typeof this.start) {
+      throw TypeError('start must be a Number');
+    }
     if (this.start < 0) {
       throw new Error('start must be >= zero');
     }
