@@ -294,6 +294,9 @@ IncomingMessage.prototype._addHeaderLine = function(field, value) {
     case 'cookie':
     case 'pragma':
     case 'link':
+    case 'www-authenticate':
+    case 'sec-websocket-extensions':
+    case 'sec-websocket-protocol':
       if (field in dest) {
         dest[field] += ', ' + value;
       } else {
@@ -1122,13 +1125,23 @@ ClientRequest.prototype.onSocket = function(socket) {
     // Setup "drain" propogation.
     httpSocketSetup(socket);
 
+    var freeParser = function() {
+      if (parser) {
+        parsers.free(parser);
+        parser = null;
+      }
+    };
+
     var errorListener = function(err) {
       debug('HTTP SOCKET ERROR: ' + err.message + '\n' + err.stack);
       req.emit('error', err);
       // For Safety. Some additional errors might fire later on
       // and we need to make sure we don't double-fire the error event.
       req._hadError = true;
-      parser.finish();
+      if (parser) {
+        parser.finish();
+        freeParser();
+      }
       socket.destroy();
     }
     socket.on('error', errorListener);
@@ -1137,6 +1150,7 @@ ClientRequest.prototype.onSocket = function(socket) {
       var ret = parser.execute(d, start, end - start);
       if (ret instanceof Error) {
         debug('parse error');
+        freeParser();
         socket.destroy(ret);
       } else if (parser.incoming && parser.incoming.upgrade) {
         var bytesParsed = ret;
@@ -1157,6 +1171,13 @@ ClientRequest.prototype.onSocket = function(socket) {
           // Got upgrade header, but have no handler.
           socket.destroy();
         }
+        freeParser();
+      } else if (parser.incoming && parser.incoming.complete &&
+                 // When the status code is 100 (Continue), the server will
+                 // send a final response after this client sends a request
+                 // body. So, we must not free the parser.
+                 parser.incoming.statusCode !== 100) {
+        freeParser();
       }
     };
 
@@ -1167,8 +1188,10 @@ ClientRequest.prototype.onSocket = function(socket) {
         req.emit('error', createHangUpError());
         req._hadError = true;
       }
-      parser.finish();
-      parsers.free(parser); // I don't know if this is necessary --Mikeal
+      if (parser) {
+        parser.finish();
+        freeParser();
+      }
       socket.destroy();
     };
 
@@ -1215,7 +1238,7 @@ ClientRequest.prototype.onSocket = function(socket) {
         return true;
       }
 
-      if (req.shouldKeepAlive && res.headers.connection !== 'keep-alive' && !req.upgraded) {
+      if (req.shouldKeepAlive && !shouldKeepAlive && !req.upgraded) {
         // Server MUST respond with Connection:keep-alive for us to enable it.
         // If we've been upgraded (via WebSockets) we also shouldn't try to
         // keep the connection open.
