@@ -481,7 +481,7 @@ function afterWrite(status, handle, req, buffer) {
     return;
   }
 
-  timers.active(this);
+  timers.active(self);
 
   self._pendingWriteReqs--;
 
@@ -570,6 +570,7 @@ Socket.prototype.connect = function(port /* [host], [cb] */) {
         // error event to the next tick.
         process.nextTick(function() {
           self.emit('error', err);
+          self.destroy();
         });
       } else {
         timers.active(self);
@@ -592,7 +593,7 @@ Socket.prototype.connect = function(port /* [host], [cb] */) {
 };
 
 
-function afterConnect(status, handle, req) {
+function afterConnect(status, handle, req, readable, writable) {
   var self = handle.socket;
 
   // callback may come after call to destroy
@@ -608,12 +609,13 @@ function afterConnect(status, handle, req) {
   self._connecting = false;
 
   if (status == 0) {
-    self.readable = self.writable = true;
+    self.readable = readable;
+    self.writable = writable;
     timers.active(self);
 
-    handle.readStart();
-
-    self.emit('connect');
+    if (self.readable) {
+      handle.readStart();
+    }
 
     if (self._connectQueue) {
       debug('Drain the connect queue');
@@ -623,6 +625,8 @@ function afterConnect(status, handle, req) {
       self._connectQueueCleanUp();
     }
 
+    self.emit('connect');
+
     if (self._flags & FLAG_SHUTDOWNQUED) {
       // end called before connected - call end now with no data
       self._flags &= ~FLAG_SHUTDOWNQUED;
