@@ -29,3 +29,16 @@ exports.totalmem = binding.getTotalMem;
 exports.cpus = binding.getCPUs;
 exports.type = binding.getOSType;
 exports.release = binding.getOSRelease;
+exports.networkInterfaces = binding.getInterfaceAddresses;
+exports.arch = function() {
+  return process.arch;
+};
+exports.platform = function() {
+  return process.platform;
+};
+
+exports.getNetworkInterfaces = function() {
+  require('util')._deprecationWarning('os',
+    'os.getNetworkInterfaces() is deprecated - use os.networkInterfaces()');
+  return exports.networkInterfaces();
+};
