@@ -96,7 +96,13 @@ if (isWindows) {
         // directories. If we've resolved a drive letter but not yet an
         // absolute path, get cwd for that drive. We're sure the device is not
         // an unc path at this points, because unc paths are always absolute.
-        path = process._cwdForDrive(resolvedDevice[0]);
+        path = process.env['=' + resolvedDevice];
+        // Verify that a drive-local cwd was found and that it actually points
+        // to our drive. If not, default to the drive's root.
+        if (!path || path.slice(0, 3).toLowerCase() !==
+            resolvedDevice.toLowerCase() + '\\') {
+          path = resolvedDevice + '\\';
+        }
       }
 
       // Skip empty and invalid entries
@@ -402,7 +408,7 @@ exports.extname = function(path) {
 
 
 exports.exists = function(path, callback) {
-  process.binding('fs').stat(path, function(err, stats) {
+  process.binding('fs').stat(_makeLong(path), function(err, stats) {
     if (callback) callback(err ? false : true);
   });
 };
@@ -410,7 +416,7 @@ exports.exists = function(path, callback) {
 
 exports.existsSync = function(path) {
   try {
-    process.binding('fs').stat(path);
+    process.binding('fs').stat(_makeLong(path));
     return true;
   } catch (e) {
     return false;
@@ -418,8 +424,13 @@ exports.existsSync = function(path) {
 };
 
 
-exports._makeLong = isWindows ? 
+var _makeLong = exports._makeLong = isWindows ?
   function(path) {
+    path = "" + path;
+    if (!path) {
+      return "";
+    }
+
     var resolvedPath = exports.resolve(path);
 
     if (resolvedPath.match(/^[a-zA-Z]\:\\/)) {
