@@ -523,8 +523,13 @@ REPLServer.prototype.complete = function(line, callback) {
             }
             // works for non-objects
             try {
-              var p = Object.getPrototypeOf(obj);
               var sentinel = 5;
+              var p;
+              if (typeof obj == 'object') {
+                p = Object.getPrototypeOf(obj);
+              } else {
+                p = obj.constructor ? obj.constructor.prototype : null;
+              }
               while (p !== null) {
                 memberGroups.push(Object.getOwnPropertyNames(p));
                 p = Object.getPrototypeOf(p);
