@@ -27,6 +27,14 @@ var stream = require('stream');
 var END_OF_FILE = 42;
 var assert = require('assert').ok;
 
+// Allow {CLIENT_RENEG_LIMIT} client-initiated session renegotiations
+// every {CLIENT_RENEG_WINDOW} seconds. An error event is emitted if more
+// renegotations are seen. The settings are applied to all remote client
+// connections.
+exports.CLIENT_RENEG_LIMIT  = 3;
+exports.CLIENT_RENEG_WINDOW = 600;
+
+
 var debug;
 if (process.env.NODE_DEBUG && /tls/.test(process.env.NODE_DEBUG)) {
   debug = function(a) { console.error('TLS:', a); };
@@ -542,6 +550,37 @@ EncryptedStream.prototype._pusher = function(pool, offset, length) {
 };
 
 
+function onhandshakestart() {
+  debug('onhandshakestart');
+
+  var self = this, ssl = this.ssl;
+  ssl.handshakes++;
+
+  if (ssl.handshakes === 1) {
+    function timeout() {
+      ssl.handshakes = 0;
+      ssl.timer = null;
+    }
+    ssl.timer = setTimeout(timeout, exports.CLIENT_RENEG_WINDOW * 1000);
+  }
+  else if (ssl.handshakes >= exports.CLIENT_RENEG_LIMIT) {
+    // Defer the error event to the next tick. We're being called from OpenSSL's
+    // state machine and OpenSSL is not re-entrant. We cannot allow the user's
+    // callback to destroy the connection right now, it would crash and burn.
+    process.nextTick(function() {
+      var err = new Error('TLS session renegotiation attack detected.');
+      if (self.cleartext) self.cleartext.emit('error', err);
+    });
+  }
+}
+
+
+function onhandshakedone() {
+  // for future use
+  debug('onhandshakedone');
+}
+
+
 /**
  * Provides a pair of streams to do encrypted communication.
  */
@@ -588,6 +627,13 @@ function SecurePair(credentials, isServer, requestCert, rejectUnauthorized,
       this._isServer ? this._requestCert : options.servername,
       this._rejectUnauthorized);
 
+  if (this._isServer) {
+    this.ssl.onhandshakestart = onhandshakestart.bind(this);
+    this.ssl.onhandshakedone = onhandshakedone.bind(this);
+    this.ssl.handshakes = 0;
+    this.ssl.timer = null;
+  }
+
   if (process.features.tls_sni) {
     if (this._isServer && options.SNICallback) {
       this.ssl.setSNICallback(options.SNICallback);
@@ -727,6 +773,12 @@ SecurePair.prototype.destroy = function() {
 
   if (!this._doneFlag) {
     this._doneFlag = true;
+
+    if (this.ssl.timer) {
+      clearTimeout(this.ssl.timer);
+      this.ssl.timer = null;
+    }
+
     this.ssl.error = null;
     this.ssl.close();
     this.ssl = null;
@@ -1030,7 +1082,7 @@ exports.connect = function(port /* host, options, cb */) {
     }
   }
 
-  var socket = new net.Stream();
+  var socket = options.socket ? options.socket : new net.Stream();
 
   var sslcontext = crypto.createCredentials(options);
 
@@ -1050,7 +1102,9 @@ exports.connect = function(port /* host, options, cb */) {
     cleartext.on('secureConnect', cb);
   }
 
-  socket.connect(port, host);
+  if (!options.socket) {
+    socket.connect(port, host);
+  }
 
   pair.on('secure', function() {
     var verifyError = pair.ssl.verifyError();
