@@ -19,6 +19,8 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+var punycode = require('punycode');
+
 exports.parse = urlParse;
 exports.resolve = urlResolve;
 exports.resolveObject = urlResolveObject;
@@ -28,7 +30,7 @@ exports.format = urlFormat;
 
 // define these here so at least they only have to be
 // compiled once on the first module load.
-var protocolPattern = /^([a-z0-9]+:)/i,
+var protocolPattern = /^([a-z0-9.+-]+:)/i,
     portPattern = /:[0-9]+$/,
     // RFC 2396: characters reserved for delimiting URLs.
     delims = ['<', '>', '"', '`', ' ', '\r', '\n', '\t'],
@@ -86,6 +88,10 @@ var protocolPattern = /^([a-z0-9]+:)/i,
 function urlParse(url, parseQueryString, slashesDenoteHost) {
   if (url && typeof(url) === 'object' && url.href) return url;
 
+  if (typeof url !== 'string') {
+    throw new TypeError("Parameter 'url' must be a string, not " + typeof url);
+  }
+
   var out = {},
       rest = url;
 
@@ -163,7 +169,6 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
 
     // pull out port.
     var p = parseHost(out.host);
-    if (out.auth) out.host = out.auth + '@' + out.host;
     var keys = Object.keys(p);
     for (var i = 0, l = keys.length; i < l; i++) {
       var key = keys[i];
@@ -183,26 +188,53 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
         var part = hostparts[i];
         if (!part) continue;
         if (!part.match(hostnamePartPattern)) {
-          var validParts = hostparts.slice(0, i);
-          var notHost = hostparts.slice(i + 1);
-          var bit = part.match(hostnamePartStart);
-          if (bit) {
-            validParts.push(bit[1]);
-            notHost.unshift(bit[2]);
+          var newpart = '';
+          for (var j = 0, k = part.length; j < k; j++) {
+            if (part.charCodeAt(j) > 127) {
+              // we replace non-ASCII char with a temporary placeholder
+              // we need this to make sure size of hostname is not
+              // broken by replacing non-ASCII by nothing
+              newpart += 'x';
+            } else {
+              newpart += part[j];
+            }
           }
-          if (notHost.length) {
-            rest = '/' + notHost.join('.') + rest
+          // we test again with ASCII char only
+          if (!newpart.match(hostnamePartPattern)) {
+            var validParts = hostparts.slice(0, i);
+            var notHost = hostparts.slice(i + 1);
+            var bit = part.match(hostnamePartStart);
+            if (bit) {
+              validParts.push(bit[1]);
+              notHost.unshift(bit[2]);
+            }
+            if (notHost.length) {
+              rest = '/' + notHost.join('.') + rest;
+            }
+            out.hostname = validParts.join('.');
+            break;
           }
-          out.hostname = validParts.join('.');
-          break;
         }
       }
     }
+
     // hostnames are always lower case.
     out.hostname = out.hostname.toLowerCase();
 
-    out.host = ((out.auth) ? out.auth + '@' : '') +
-        (out.hostname || '') +
+    // IDNA Support: Returns a puny coded representation of "domain".
+    // It only converts the part of the domain name that
+    // has non ASCII characters. I.e. it dosent matter if
+    // you call it with a domain that already is in ASCII.
+    var domainArray = out.hostname.split('.');
+    var newOut = [];
+    for (var i = 0; i < domainArray.length; ++i) {
+      var s = domainArray[i];
+      newOut.push(s.match(/[^A-Za-z0-9_-]/) ?
+          'xn--' + punycode.encode(s) : s);
+    }
+    out.hostname = newOut.join('.');
+
+    out.host = (out.hostname || '') +
         ((out.port) ? ':' + out.port : '');
     out.href += out.host;
   }
@@ -261,9 +293,14 @@ function urlParse(url, parseQueryString, slashesDenoteHost) {
     out.pathname = '/';
   }
 
+  //to support http.request
+  if (out.pathname || out.search) {
+    out.path = (out.pathname ? out.pathname : '') +
+               (out.search ? out.search : '');
+  }
+
   // finally, reconstruct the href based on what has been validated.
   out.href = urlFormat(out);
-
   return out;
 }
 
@@ -275,20 +312,20 @@ function urlFormat(obj) {
   // to clean up potentially wonky urls.
   if (typeof(obj) === 'string') obj = urlParse(obj);
 
-  var auth = obj.auth;
+  var auth = obj.auth || '';
   if (auth) {
     auth = auth.split('@').join('%40');
     for (var i = 0, l = nonAuthChars.length; i < l; i++) {
       var nAC = nonAuthChars[i];
       auth = auth.split(nAC).join(encodeURIComponent(nAC));
     }
+    auth += '@';
   }
 
   var protocol = obj.protocol || '',
-      host = (obj.host !== undefined) ? obj.host :
+      host = (obj.host !== undefined) ? auth + obj.host :
           obj.hostname !== undefined ? (
-              (auth ? auth + '@' : '') +
-              obj.hostname +
+              auth + obj.hostname +
               (obj.port ? ':' + obj.port : '')
           ) :
           false,
@@ -332,11 +369,20 @@ function urlResolveObject(source, relative) {
   // hash is always overridden, no matter what.
   source.hash = relative.hash;
 
-  if (relative.href === '') return source;
+  if (relative.href === '') {
+    source.href = urlFormat(source);
+    return source;
+  }
 
   // hrefs like //foo/bar always cut to the protocol.
   if (relative.slashes && !relative.protocol) {
     relative.protocol = source.protocol;
+    //urlParse appends trailing / to urls like http://www.example.com
+    if (slashedProtocol[relative.protocol] &&
+        relative.hostname && !relative.pathname) {
+      relative.path = relative.pathname = '/';
+    }
+    relative.href = urlFormat(relative);
     return relative;
   }
 
@@ -349,14 +395,16 @@ function urlResolveObject(source, relative) {
     // if it is file:, then the host is dropped,
     // because that's known to be hostless.
     // anything else is assumed to be absolute.
-
-    if (!slashedProtocol[relative.protocol]) return relative;
-
+    if (!slashedProtocol[relative.protocol]) {
+      relative.href = urlFormat(relative);
+      return relative;
+    }
     source.protocol = relative.protocol;
     if (!relative.host && !hostlessProtocol[relative.protocol]) {
       var relPath = (relative.pathname || '').split('/');
       while (relPath.length && !(relative.host = relPath.shift()));
       if (!relative.host) relative.host = '';
+      if (!relative.hostname) relative.hostname = '';
       if (relPath[0] !== '') relPath.unshift('');
       if (relPath.length < 2) relPath.unshift('');
       relative.pathname = relPath.join('/');
@@ -365,9 +413,16 @@ function urlResolveObject(source, relative) {
     source.search = relative.search;
     source.query = relative.query;
     source.host = relative.host || '';
-    delete source.auth;
-    delete source.hostname;
+    source.auth = relative.auth;
+    source.hostname = relative.hostname || relative.host;
     source.port = relative.port;
+    //to support http.request
+    if (source.pathname !== undefined || source.search !== undefined) {
+      source.path = (source.pathname ? source.pathname : '') +
+                    (source.search ? source.search : '');
+    }
+    source.slashes = source.slashes || relative.slashes;
+    source.href = urlFormat(source);
     return source;
   }
 
@@ -382,8 +437,7 @@ function urlResolveObject(source, relative) {
       srcPath = source.pathname && source.pathname.split('/') || [],
       relPath = relative.pathname && relative.pathname.split('/') || [],
       psychotic = source.protocol &&
-          !slashedProtocol[source.protocol] &&
-          source.host !== undefined;
+          !slashedProtocol[source.protocol];
 
   // if the url is a non-slashed url, then relative
   // links like ../.. should be able
@@ -393,17 +447,14 @@ function urlResolveObject(source, relative) {
   if (psychotic) {
 
     delete source.hostname;
-    delete source.auth;
     delete source.port;
     if (source.host) {
       if (srcPath[0] === '') srcPath[0] = source.host;
       else srcPath.unshift(source.host);
     }
     delete source.host;
-
     if (relative.protocol) {
       delete relative.hostname;
-      delete relative.auth;
       delete relative.port;
       if (relative.host) {
         if (relPath[0] === '') relPath[0] = relative.host;
@@ -418,6 +469,8 @@ function urlResolveObject(source, relative) {
     // it's absolute.
     source.host = (relative.host || relative.host === '') ?
                       relative.host : source.host;
+    source.hostname = (relative.hostname || relative.hostname === '') ?
+                      relative.hostname : source.hostname;
     source.search = relative.search;
     source.query = relative.query;
     srcPath = relPath;
@@ -435,19 +488,40 @@ function urlResolveObject(source, relative) {
     // like href='?foo'.
     // Put this after the other two cases because it simplifies the booleans
     if (psychotic) {
-      source.host = srcPath.shift();
+      source.hostname = source.host = srcPath.shift();
+      //occationaly the auth can get stuck only in host
+      //this especialy happens in cases like
+      //url.resolveObject('mailto:local1@domain1', 'local2@domain2')
+      var authInHost = source.host && source.host.indexOf('@') > 0 ?
+                       source.host.split('@') : false;
+      if (authInHost) {
+        source.auth = authInHost.shift();
+        source.host = source.hostname = authInHost.shift();
+      }
     }
     source.search = relative.search;
     source.query = relative.query;
+    //to support http.request
+    if (source.pathname !== undefined || source.search !== undefined) {
+      source.path = (source.pathname ? source.pathname : '') +
+                    (source.search ? source.search : '');
+    }
+    source.href = urlFormat(source);
     return source;
   }
   if (!srcPath.length) {
     // no path at all.  easy.
     // we've already handled the other stuff above.
     delete source.pathname;
+    //to support http.request
+    if (!source.search) {
+      source.path = '/' + source.search;
+    } else {
+      delete source.path;
+    }
+    source.href = urlFormat(source);
     return source;
   }
-
   // if a url ENDs in . or .., then it must get a trailing slash.
   // however, if it ends in anything else non-slashy,
   // then it must NOT get a trailing slash.
@@ -493,7 +567,17 @@ function urlResolveObject(source, relative) {
 
   // put the host back
   if (psychotic) {
-    source.host = isAbsolute ? '' : srcPath.shift();
+    source.hostname = source.host = isAbsolute ? '' :
+                                    srcPath.length ? srcPath.shift() : '';
+    //occationaly the auth can get stuck only in host
+    //this especialy happens in cases like
+    //url.resolveObject('mailto:local1@domain1', 'local2@domain2')
+    var authInHost = source.host && source.host.indexOf('@') > 0 ?
+                     source.host.split('@') : false;
+    if (authInHost) {
+      source.auth = authInHost.shift();
+      source.host = source.hostname = authInHost.shift();
+    }
   }
 
   mustEndAbs = mustEndAbs || (source.host && srcPath.length);
@@ -503,8 +587,14 @@ function urlResolveObject(source, relative) {
   }
 
   source.pathname = srcPath.join('/');
-
-
+  //to support request.http
+  if (source.pathname !== undefined || source.search !== undefined) {
+    source.path = (source.pathname ? source.pathname : '') +
+                  (source.search ? source.search : '');
+  }
+  source.auth = relative.auth || source.auth;
+  source.slashes = source.slashes || relative.slashes;
+  source.href = urlFormat(source);
   return source;
 }
 
