@@ -19,75 +19,59 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-#include <node.h>
-#include <node_file.h>
-#include <node_buffer.h>
-#include <node_stat_watcher.h>
+#include "node.h"
+#include "node_file.h"
+#include "node_buffer.h"
+#ifdef __POSIX__
+# include "node_stat_watcher.h"
+#endif
+#include "req_wrap.h"
 
+#include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <dirent.h>
-#include <fcntl.h>
-#include <stdlib.h>
-#include <unistd.h>
 #include <assert.h>
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
 
-#ifdef __MINGW32__
+#if defined(__MINGW32__) || defined(_MSC_VER)
+# include <io.h>
 # include <platform_win32.h>
 #endif
 
-/* used for readlink, AIX doesn't provide it */
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
-/* HACK to use pread/pwrite from eio because MINGW32 doesn't have it */
-/* TODO fixme */
-#ifdef __MINGW32__
-# define pread  eio__pread
-# define pwrite eio__pwrite
-#endif
 
 namespace node {
 
 using namespace v8;
 
-#define ARRAY_SIZE(a) (sizeof(a) / sizeof(*(a)))
 #define MIN(a,b) ((a) < (b) ? (a) : (b))
 #define THROW_BAD_ARGS \
   ThrowException(Exception::TypeError(String::New("Bad argument")))
 
+typedef class ReqWrap<uv_fs_t> FSReqWrap;
+
 static Persistent<String> encoding_symbol;
 static Persistent<String> errno_symbol;
 static Persistent<String> buf_symbol;
-
-// Buffer for readlink()  and other misc callers; keep this scoped at
-// file-level rather than method-level to avoid excess stack usage.
-// Not used on windows atm
-#ifdef __POSIX__
-  static char getbuf[PATH_MAX + 1];
-#endif
+static Persistent<String> oncomplete_sym;
 
 
-static inline bool SetCloseOnExec(int fd) {
-#ifdef __POSIX__
-  return (fcntl(fd, F_SETFD, FD_CLOEXEC) != -1);
-#else // __MINGW32__
-  return SetHandleInformation(reinterpret_cast<HANDLE>(_get_osfhandle(fd)),
-                              HANDLE_FLAG_INHERIT, 0) != 0;
-#endif
+#ifdef _LARGEFILE_SOURCE
+static inline int IsInt64(double x) {
+  return x == static_cast<double>(static_cast<int64_t>(x));
 }
+#endif
 
 
-static int After(eio_req *req) {
+static void After(uv_fs_t *req) {
   HandleScope scope;
 
-  Persistent<Function> *callback = cb_unwrap(req->data);
-
-  ev_unref(EV_DEFAULT_UC);
+  FSReqWrap* req_wrap = (FSReqWrap*) req->data;
+  assert(&req_wrap->req_ == req);
+  Local<Value> callback_v = req_wrap->object_->Get(oncomplete_sym);
+  assert(callback_v->IsFunction());
+  Local<Function> callback = Local<Function>::Cast(callback_v);
 
   // there is always at least one argument. "error"
   int argc = 1;
@@ -100,10 +84,14 @@ static int After(eio_req *req) {
   // for a success, which is possible.
   if (req->result == -1) {
     // If the request doesn't have a path parameter set.
-    if (!req->ptr1) {
-      argv[0] = ErrnoException(req->errorno);
+
+    if (!req->path) {
+      argv[0] = UVException(req->errorno);
     } else {
-      argv[0] = ErrnoException(req->errorno, NULL, "", static_cast<const char*>(req->ptr1));
+      argv[0] = UVException(req->errorno,
+                            NULL,
+                            NULL,
+                            static_cast<const char*>(req->path));
     }
   } else {
     // error value is empty or null for non-error.
@@ -112,56 +100,62 @@ static int After(eio_req *req) {
     // All have at least two args now.
     argc = 2;
 
-    switch (req->type) {
+    switch (req->fs_type) {
       // These all have no data to pass.
-      case EIO_CLOSE:
-      case EIO_RENAME:
-      case EIO_UNLINK:
-      case EIO_RMDIR:
-      case EIO_MKDIR:
-      case EIO_FTRUNCATE:
-      case EIO_FSYNC:
-      case EIO_FDATASYNC:
-      case EIO_LINK:
-      case EIO_SYMLINK:
-      case EIO_CHMOD:
-      case EIO_CHOWN:
+      case UV_FS_CLOSE:
+      case UV_FS_RENAME:
+      case UV_FS_UNLINK:
+      case UV_FS_RMDIR:
+      case UV_FS_MKDIR:
+      case UV_FS_FTRUNCATE:
+      case UV_FS_FSYNC:
+      case UV_FS_FDATASYNC:
+      case UV_FS_LINK:
+      case UV_FS_SYMLINK:
+      case UV_FS_CHMOD:
+      case UV_FS_FCHMOD:
+      case UV_FS_CHOWN:
+      case UV_FS_FCHOWN:
         // These, however, don't.
         argc = 1;
         break;
 
-      case EIO_OPEN:
-        SetCloseOnExec(req->result);
+      case UV_FS_UTIME:
+      case UV_FS_FUTIME:
+        argc = 0;
+        break;
+
+      case UV_FS_OPEN:
         /* pass thru */
-      case EIO_SENDFILE:
+      case UV_FS_SENDFILE:
         argv[1] = Integer::New(req->result);
         break;
 
-      case EIO_WRITE:
+      case UV_FS_WRITE:
         argv[1] = Integer::New(req->result);
         break;
 
-      case EIO_STAT:
-      case EIO_LSTAT:
-      case EIO_FSTAT:
+      case UV_FS_STAT:
+      case UV_FS_LSTAT:
+      case UV_FS_FSTAT:
         {
-          NODE_STAT_STRUCT *s = reinterpret_cast<NODE_STAT_STRUCT*>(req->ptr2);
+          NODE_STAT_STRUCT *s = reinterpret_cast<NODE_STAT_STRUCT*>(req->ptr);
           argv[1] = BuildStatsObject(s);
         }
         break;
 
-      case EIO_READLINK:
-        argv[1] = String::New(static_cast<char*>(req->ptr2), req->result);
+      case UV_FS_READLINK:
+        argv[1] = String::New(static_cast<char*>(req->ptr));
         break;
 
-      case EIO_READ:
+      case UV_FS_READ:
         // Buffer interface
         argv[1] = Integer::New(req->result);
         break;
 
-      case EIO_READDIR:
+      case UV_FS_READDIR:
         {
-          char *namebuf = static_cast<char*>(req->ptr2);
+          char *namebuf = static_cast<char*>(req->ptr);
           int nnames = req->result;
 
           Local<Array> names = Array::New(nnames);
@@ -189,24 +183,49 @@ static int After(eio_req *req) {
 
   TryCatch try_catch;
 
-  (*callback)->Call(v8::Context::GetCurrent()->Global(), argc, argv);
+  callback->Call(req_wrap->object_, argc, argv);
 
   if (try_catch.HasCaught()) {
     FatalException(try_catch);
   }
 
-  // Dispose of the persistent handle
-  cb_destroy(callback);
-
-  return 0;
+  uv_fs_req_cleanup(&req_wrap->req_);
+  delete req_wrap;
 }
 
+// This struct is only used on sync fs calls.
+// For async calls FSReqWrap is used.
+struct fs_req_wrap {
+  fs_req_wrap() {}
+  ~fs_req_wrap() { uv_fs_req_cleanup(&req); }
+  // Ensure that copy ctor and assignment operator are not used.
+  fs_req_wrap(const fs_req_wrap& req);
+  fs_req_wrap& operator=(const fs_req_wrap& req);
+  uv_fs_t req;
+};
+
+
 #define ASYNC_CALL(func, callback, ...)                           \
-  eio_req *req = eio_##func(__VA_ARGS__, EIO_PRI_DEFAULT, After,  \
-    cb_persist(callback));                                        \
-  assert(req);                                                    \
-  ev_ref(EV_DEFAULT_UC);                                          \
-  return Undefined();
+  FSReqWrap* req_wrap = new FSReqWrap();                          \
+  int r = uv_fs_##func(uv_default_loop(), &req_wrap->req_,        \
+      __VA_ARGS__, After);                                        \
+  assert(r == 0);                                                 \
+  req_wrap->object_->Set(oncomplete_sym, callback);               \
+  req_wrap->Dispatched();                                         \
+  return scope.Close(req_wrap->object_);
+
+#define SYNC_CALL(func, path, ...)                                \
+  fs_req_wrap req_wrap;                                           \
+  int result = uv_fs_##func(uv_default_loop(), &req_wrap.req, __VA_ARGS__, NULL); \
+  if (result < 0) {                                               \
+    int code = uv_last_error(uv_default_loop()).code;             \
+    return ThrowException(UVException(code, #func, "", path));    \
+  }
+
+#define SYNC_REQ req_wrap.req
+
+#define SYNC_RESULT result
+
 
 static Handle<Value> Close(const Arguments& args) {
   HandleScope scope;
@@ -220,8 +239,7 @@ static Handle<Value> Close(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(close, args[1], fd)
   } else {
-    int ret = close(fd);
-    if (ret != 0) return ThrowException(ErrnoException(errno));
+    SYNC_CALL(close, 0, fd)
     return Undefined();
   }
 }
@@ -321,14 +339,11 @@ static Handle<Value> Stat(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(stat, args[1], *path)
   } else {
-    NODE_STAT_STRUCT s;
-    int ret = NODE_STAT(*path, &s);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
-    return scope.Close(BuildStatsObject(&s));
+    SYNC_CALL(stat, *path, *path)
+    return scope.Close(BuildStatsObject((NODE_STAT_STRUCT*)SYNC_REQ.ptr));
   }
 }
 
-#ifdef __POSIX__
 static Handle<Value> LStat(const Arguments& args) {
   HandleScope scope;
 
@@ -341,13 +356,10 @@ static Handle<Value> LStat(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(lstat, args[1], *path)
   } else {
-    NODE_STAT_STRUCT s;
-    int ret = lstat(*path, &s);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
-    return scope.Close(BuildStatsObject(&s));
+    SYNC_CALL(lstat, *path, *path)
+    return scope.Close(BuildStatsObject((NODE_STAT_STRUCT*)SYNC_REQ.ptr));
   }
 }
-#endif // __POSIX__
 
 static Handle<Value> FStat(const Arguments& args) {
   HandleScope scope;
@@ -361,14 +373,11 @@ static Handle<Value> FStat(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(fstat, args[1], fd)
   } else {
-    NODE_STAT_STRUCT s;
-    int ret = NODE_FSTAT(fd, &s);
-    if (ret != 0) return ThrowException(ErrnoException(errno));
-    return scope.Close(BuildStatsObject(&s));
+    SYNC_CALL(fstat, 0, fd)
+    return scope.Close(BuildStatsObject((NODE_STAT_STRUCT*)SYNC_REQ.ptr));
   }
 }
 
-#ifdef __POSIX__
 static Handle<Value> Symlink(const Arguments& args) {
   HandleScope scope;
 
@@ -378,18 +387,23 @@ static Handle<Value> Symlink(const Arguments& args) {
 
   String::Utf8Value dest(args[0]->ToString());
   String::Utf8Value path(args[1]->ToString());
+  int flags = 0;
 
-  if (args[2]->IsFunction()) {
-    ASYNC_CALL(symlink, args[2], *dest, *path)
+  if (args[2]->IsString()) {
+    String::Utf8Value mode(args[2]->ToString());
+    if (memcmp(*mode, "dir\0", 4) == 0) {
+      flags |= UV_FS_SYMLINK_DIR;
+    }
+  }
+
+  if (args[3]->IsFunction()) {
+    ASYNC_CALL(symlink, args[3], *dest, *path, flags)
   } else {
-    int ret = symlink(*dest, *path);
-    if (ret != 0) return ThrowException(ErrnoException(errno));
+    SYNC_CALL(symlink, *path, *dest, *path, flags)
     return Undefined();
   }
 }
-#endif // __POSIX__
 
-#ifdef __POSIX__
 static Handle<Value> Link(const Arguments& args) {
   HandleScope scope;
 
@@ -403,14 +417,11 @@ static Handle<Value> Link(const Arguments& args) {
   if (args[2]->IsFunction()) {
     ASYNC_CALL(link, args[2], *orig_path, *new_path)
   } else {
-    int ret = link(*orig_path, *new_path);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *orig_path));
+    SYNC_CALL(link, *orig_path, *orig_path, *new_path)
     return Undefined();
   }
 }
-#endif // __POSIX__
 
-#ifdef __POSIX__
 static Handle<Value> ReadLink(const Arguments& args) {
   HandleScope scope;
 
@@ -423,13 +434,10 @@ static Handle<Value> ReadLink(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(readlink, args[1], *path)
   } else {
-    ssize_t bz = readlink(*path, getbuf, ARRAY_SIZE(getbuf) - 1);
-    if (bz == -1) return ThrowException(ErrnoException(errno, NULL, "", *path));
-    getbuf[ARRAY_SIZE(getbuf) - 1] = '\0';
-    return scope.Close(String::New(getbuf, bz));
+    SYNC_CALL(readlink, *path, *path)
+    return scope.Close(String::New((char*)SYNC_REQ.ptr));
   }
 }
-#endif // __POSIX__
 
 static Handle<Value> Rename(const Arguments& args) {
   HandleScope scope;
@@ -444,12 +452,25 @@ static Handle<Value> Rename(const Arguments& args) {
   if (args[2]->IsFunction()) {
     ASYNC_CALL(rename, args[2], *old_path, *new_path)
   } else {
-    int ret = rename(*old_path, *new_path);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *old_path));
+    SYNC_CALL(rename, *old_path, *old_path, *new_path)
     return Undefined();
   }
 }
 
+#ifndef _LARGEFILE_SOURCE
+#define ASSERT_TRUNCATE_LENGTH(a) \
+  if (!(a)->IsUndefined() && !(a)->IsNull() && !(a)->IsUint32()) { \
+    return ThrowException(Exception::TypeError(String::New("Not an integer"))); \
+  }
+#define GET_TRUNCATE_LENGTH(a) ((a)->Uint32Value())
+#else
+#define ASSERT_TRUNCATE_LENGTH(a) \
+  if (!(a)->IsUndefined() && !(a)->IsNull() && !IsInt64((a)->NumberValue())) { \
+    return ThrowException(Exception::TypeError(String::New("Not an integer"))); \
+  }
+#define GET_TRUNCATE_LENGTH(a) ((a)->IntegerValue())
+#endif
+
 static Handle<Value> Truncate(const Arguments& args) {
   HandleScope scope;
 
@@ -458,13 +479,14 @@ static Handle<Value> Truncate(const Arguments& args) {
   }
 
   int fd = args[0]->Int32Value();
-  off_t len = args[1]->Uint32Value();
+
+  ASSERT_TRUNCATE_LENGTH(args[1]);
+  off_t len = GET_TRUNCATE_LENGTH(args[1]);
 
   if (args[2]->IsFunction()) {
     ASYNC_CALL(ftruncate, args[2], fd, len)
   } else {
-    int ret = ftruncate(fd, len);
-    if (ret != 0) return ThrowException(ErrnoException(errno));
+    SYNC_CALL(ftruncate, 0, fd, len)
     return Undefined();
   }
 }
@@ -481,14 +503,7 @@ static Handle<Value> Fdatasync(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(fdatasync, args[1], fd)
   } else {
-#if HAVE_FDATASYNC
-    int ret = fdatasync(fd);
-#elif defined(__MINGW32__)
-    int ret = FlushFileBuffers((HANDLE)_get_osfhandle(fd)) ? 0 : -1;
-#else
-    int ret = fsync(fd);
-#endif
-    if (ret != 0) return ThrowException(ErrnoException(errno));
+    SYNC_CALL(fdatasync, 0, fd)
     return Undefined();
   }
 }
@@ -505,12 +520,7 @@ static Handle<Value> Fsync(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(fsync, args[1], fd)
   } else {
-#ifdef __MINGW32__
-    int ret = FlushFileBuffers((HANDLE)_get_osfhandle(fd)) ? 0 : -1;
-#else
-    int ret = fsync(fd);
-#endif
-    if (ret != 0) return ThrowException(ErrnoException(errno));
+    SYNC_CALL(fsync, 0, fd)
     return Undefined();
   }
 }
@@ -527,8 +537,7 @@ static Handle<Value> Unlink(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(unlink, args[1], *path)
   } else {
-    int ret = unlink(*path);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(unlink, *path, *path)
     return Undefined();
   }
 }
@@ -545,8 +554,7 @@ static Handle<Value> RMDir(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(rmdir, args[1], *path)
   } else {
-    int ret = rmdir(*path);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(rmdir, *path, *path)
     return Undefined();
   }
 }
@@ -559,17 +567,12 @@ static Handle<Value> MKDir(const Arguments& args) {
   }
 
   String::Utf8Value path(args[0]->ToString());
-  mode_t mode = static_cast<mode_t>(args[1]->Int32Value());
+  int mode = static_cast<int>(args[1]->Int32Value());
 
   if (args[2]->IsFunction()) {
     ASYNC_CALL(mkdir, args[2], *path, mode)
   } else {
-#ifdef __MINGW32__
-    int ret = mkdir(*path);
-#else
-    int ret = mkdir(*path, mode);
-#endif
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(mkdir, *path, *path, mode)
     return Undefined();
   }
 }
@@ -593,10 +596,8 @@ static Handle<Value> SendFile(const Arguments& args) {
   if (args[4]->IsFunction()) {
     ASYNC_CALL(sendfile, args[4], out_fd, in_fd, in_offset, length)
   } else {
-    ssize_t sent = eio_sendfile_sync (out_fd, in_fd, in_offset, length);
-    // XXX is this the right errno to use?
-    if (sent < 0) return ThrowException(ErrnoException(errno));
-    return scope.Close(Integer::New(sent));
+    SYNC_CALL(sendfile, 0, out_fd, in_fd, in_offset, length)
+    return scope.Close(Integer::New(SYNC_RESULT));
   }
 }
 
@@ -612,27 +613,25 @@ static Handle<Value> ReadDir(const Arguments& args) {
   if (args[1]->IsFunction()) {
     ASYNC_CALL(readdir, args[1], *path, 0 /*flags*/)
   } else {
-    DIR *dir = opendir(*path);
-    if (!dir) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(readdir, *path, *path, 0 /*flags*/)
 
-    struct dirent *ent;
+    char *namebuf = static_cast<char*>(SYNC_REQ.ptr);
+    int nnames = req_wrap.req.result;
+    Local<Array> names = Array::New(nnames);
 
-    Local<Array> files = Array::New();
-    char *name;
-    int i = 0;
-
-    while ((ent = readdir(dir))) {
-      name = ent->d_name;
-
-      if (name[0] != '.' || (name[1] && (name[1] != '.' || name[2]))) {
-        files->Set(Integer::New(i), String::New(name));
-        i++;
-      }
+    for (int i = 0; i < nnames; i++) {
+      Local<String> name = String::New(namebuf);
+      names->Set(Integer::New(i), name);
+#ifndef NDEBUG
+      namebuf += strlen(namebuf);
+      assert(*namebuf == '\0');
+      namebuf += 1;
+#else
+      namebuf += strlen(namebuf) + 1;
+#endif
     }
 
-    closedir(dir);
-
-    return scope.Close(files);
+    return scope.Close(names);
   }
 }
 
@@ -648,19 +647,30 @@ static Handle<Value> Open(const Arguments& args) {
 
   String::Utf8Value path(args[0]->ToString());
   int flags = args[1]->Int32Value();
-  mode_t mode = static_cast<mode_t>(args[2]->Int32Value());
+  int mode = static_cast<int>(args[2]->Int32Value());
 
   if (args[3]->IsFunction()) {
     ASYNC_CALL(open, args[3], *path, flags, mode)
   } else {
-    int fd = open(*path, flags, mode);
-    SetCloseOnExec(fd);
-    if (fd < 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(open, *path, *path, flags, mode)
+    int fd = SYNC_RESULT;
     return scope.Close(Integer::New(fd));
   }
 }
 
-#define GET_OFFSET(a) (a)->IsInt32() ? (a)->IntegerValue() : -1;
+#ifndef _LARGEFILE_SOURCE
+#define ASSERT_OFFSET(a) \
+  if (!(a)->IsUndefined() && !(a)->IsNull() && !(a)->IsInt32()) { \
+    return ThrowException(Exception::TypeError(String::New("Not an integer"))); \
+  }
+#define GET_OFFSET(a) ((a)->IsNumber() ? (a)->Int32Value() : -1)
+#else
+#define ASSERT_OFFSET(a) \
+  if (!(a)->IsUndefined() && !(a)->IsNull() && !IsInt64((a)->NumberValue())) { \
+    return ThrowException(Exception::TypeError(String::New("Not an integer"))); \
+  }
+#define GET_OFFSET(a) ((a)->IsNumber() ? (a)->IntegerValue() : -1)
+#endif
 
 // bytesWritten = write(fd, data, position, enc, callback)
 // Wrapper for write(2).
@@ -701,18 +711,17 @@ static Handle<Value> Write(const Arguments& args) {
           String::New("Length is extends beyond buffer")));
   }
 
+  ASSERT_OFFSET(args[4]);
   off_t pos = GET_OFFSET(args[4]);
 
   char * buf = (char*)buffer_data + off;
   Local<Value> cb = args[5];
 
   if (cb->IsFunction()) {
-
     ASYNC_CALL(write, cb, fd, buf, len, pos)
   } else {
-    ssize_t written = pos < 0 ? write(fd, buf, len) : pwrite(fd, buf, len, pos);
-    if (written < 0) return ThrowException(ErrnoException(errno, "write"));
-    return scope.Close(Integer::New(written));
+    SYNC_CALL(write, 0, fd, buf, len, pos)
+    return scope.Close(Integer::New(SYNC_RESULT));
   }
 }
 
@@ -762,7 +771,7 @@ static Handle<Value> Read(const Arguments& args) {
   len = args[3]->Int32Value();
   if (off + len > buffer_length) {
     return ThrowException(Exception::Error(
-          String::New("Length is extends beyond buffer")));
+          String::New("Length extends beyond buffer")));
   }
 
   pos = GET_OFFSET(args[4]);
@@ -772,21 +781,16 @@ static Handle<Value> Read(const Arguments& args) {
   cb = args[5];
 
   if (cb->IsFunction()) {
-
     ASYNC_CALL(read, cb, fd, buf, len, pos);
   } else {
-    // SYNC
-    ssize_t ret;
-
-    ret = pos < 0 ? read(fd, buf, len) : pread(fd, buf, len, pos);
-    if (ret < 0) return ThrowException(ErrnoException(errno));
-    Local<Integer> bytesRead = Integer::New(ret);
+    SYNC_CALL(read, 0, fd, buf, len, pos)
+    Local<Integer> bytesRead = Integer::New(SYNC_RESULT);
     return scope.Close(bytesRead);
   }
 }
 
 
-/* fs.chmod(fd, mode);
+/* fs.chmod(path, mode);
  * Wrapper for chmod(1) / EIO_CHMOD
  */
 static Handle<Value> Chmod(const Arguments& args) {
@@ -796,22 +800,41 @@ static Handle<Value> Chmod(const Arguments& args) {
     return THROW_BAD_ARGS;
   }
   String::Utf8Value path(args[0]->ToString());
-  mode_t mode = static_cast<mode_t>(args[1]->Int32Value());
+  int mode = static_cast<int>(args[1]->Int32Value());
 
   if(args[2]->IsFunction()) {
     ASYNC_CALL(chmod, args[2], *path, mode);
   } else {
-    int ret = chmod(*path, mode);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(chmod, *path, *path, mode);
     return Undefined();
   }
 }
 
 
-/* fs.chown(fd, uid, gid);
+/* fs.fchmod(fd, mode);
+ * Wrapper for fchmod(1) / EIO_FCHMOD
+ */
+static Handle<Value> FChmod(const Arguments& args) {
+  HandleScope scope;
+
+  if(args.Length() < 2 || !args[0]->IsInt32() || !args[1]->IsInt32()) {
+    return THROW_BAD_ARGS;
+  }
+  int fd = args[0]->Int32Value();
+  int mode = static_cast<int>(args[1]->Int32Value());
+
+  if(args[2]->IsFunction()) {
+    ASYNC_CALL(fchmod, args[2], fd, mode);
+  } else {
+    SYNC_CALL(fchmod, 0, fd, mode);
+    return Undefined();
+  }
+}
+
+
+/* fs.chown(path, uid, gid);
  * Wrapper for chown(1) / EIO_CHOWN
  */
-#ifdef __POSIX__
 static Handle<Value> Chown(const Arguments& args) {
   HandleScope scope;
 
@@ -824,18 +847,91 @@ static Handle<Value> Chown(const Arguments& args) {
   }
 
   String::Utf8Value path(args[0]->ToString());
-  uid_t uid = static_cast<uid_t>(args[1]->Int32Value());
-  gid_t gid = static_cast<gid_t>(args[2]->Int32Value());
+  int uid = static_cast<int>(args[1]->Int32Value());
+  int gid = static_cast<int>(args[2]->Int32Value());
 
   if (args[3]->IsFunction()) {
     ASYNC_CALL(chown, args[3], *path, uid, gid);
   } else {
-    int ret = chown(*path, uid, gid);
-    if (ret != 0) return ThrowException(ErrnoException(errno, NULL, "", *path));
+    SYNC_CALL(chown, *path, *path, uid, gid);
     return Undefined();
   }
 }
-#endif // __POSIX__
+
+
+/* fs.fchown(fd, uid, gid);
+ * Wrapper for fchown(1) / EIO_FCHOWN
+ */
+static Handle<Value> FChown(const Arguments& args) {
+  HandleScope scope;
+
+  if (args.Length() < 3 || !args[0]->IsInt32()) {
+    return THROW_BAD_ARGS;
+  }
+
+  if (!args[1]->IsInt32() || !args[2]->IsInt32()) {
+    return ThrowException(Exception::Error(String::New("User and Group IDs must be an integer.")));
+  }
+
+  int fd = args[0]->Int32Value();
+  int uid = static_cast<int>(args[1]->Int32Value());
+  int gid = static_cast<int>(args[2]->Int32Value());
+
+  if (args[3]->IsFunction()) {
+    ASYNC_CALL(fchown, args[3], fd, uid, gid);
+  } else {
+    SYNC_CALL(fchown, 0, fd, uid, gid);
+    return Undefined();
+  }
+}
+
+
+static Handle<Value> UTimes(const Arguments& args) {
+  HandleScope scope;
+
+  if (args.Length() < 3
+      || !args[0]->IsString()
+      || !args[1]->IsNumber()
+      || !args[2]->IsNumber())
+  {
+    return THROW_BAD_ARGS;
+  }
+
+  const String::Utf8Value path(args[0]->ToString());
+  const double atime = static_cast<double>(args[1]->NumberValue());
+  const double mtime = static_cast<double>(args[2]->NumberValue());
+
+  if (args[3]->IsFunction()) {
+    ASYNC_CALL(utime, args[3], *path, atime, mtime);
+  } else {
+    SYNC_CALL(utime, *path, *path, atime, mtime);
+    return Undefined();
+  }
+}
+
+static Handle<Value> FUTimes(const Arguments& args) {
+  HandleScope scope;
+
+  if (args.Length() < 3
+      || !args[0]->IsInt32()
+      || !args[1]->IsNumber()
+      || !args[2]->IsNumber())
+  {
+    return THROW_BAD_ARGS;
+  }
+
+  const int fd = args[0]->Int32Value();
+  const double atime = static_cast<double>(args[1]->NumberValue());
+  const double mtime = static_cast<double>(args[2]->NumberValue());
+
+  if (args[3]->IsFunction()) {
+    ASYNC_CALL(futime, args[3], fd, atime, mtime);
+  } else {
+    SYNC_CALL(futime, 0, fd, atime, mtime);
+    return Undefined();
+  }
+}
+
 
 void File::Initialize(Handle<Object> target) {
   HandleScope scope;
@@ -852,22 +948,24 @@ void File::Initialize(Handle<Object> target) {
   NODE_SET_METHOD(target, "sendfile", SendFile);
   NODE_SET_METHOD(target, "readdir", ReadDir);
   NODE_SET_METHOD(target, "stat", Stat);
-#ifdef __POSIX__
   NODE_SET_METHOD(target, "lstat", LStat);
-#endif // __POSIX__
   NODE_SET_METHOD(target, "fstat", FStat);
-#ifdef __POSIX__
   NODE_SET_METHOD(target, "link", Link);
   NODE_SET_METHOD(target, "symlink", Symlink);
   NODE_SET_METHOD(target, "readlink", ReadLink);
-#endif // __POSIX__
   NODE_SET_METHOD(target, "unlink", Unlink);
   NODE_SET_METHOD(target, "write", Write);
 
   NODE_SET_METHOD(target, "chmod", Chmod);
-#ifdef __POSIX__
+  NODE_SET_METHOD(target, "fchmod", FChmod);
+  //NODE_SET_METHOD(target, "lchmod", LChmod);
+
   NODE_SET_METHOD(target, "chown", Chown);
-#endif // __POSIX__
+  NODE_SET_METHOD(target, "fchown", FChown);
+  //NODE_SET_METHOD(target, "lchown", LChown);
+
+  NODE_SET_METHOD(target, "utimes", UTimes);
+  NODE_SET_METHOD(target, "futimes", FUTimes);
 
   errno_symbol = NODE_PSYMBOL("errno");
   encoding_symbol = NODE_PSYMBOL("node:encoding");
@@ -881,15 +979,15 @@ void InitFs(Handle<Object> target) {
   stats_constructor_template = Persistent<FunctionTemplate>::New(stat_templ);
   target->Set(String::NewSymbol("Stats"),
                stats_constructor_template->GetFunction());
-  StatWatcher::Initialize(target);
   File::Initialize(target);
 
-#ifdef __MINGW32__
-  // Open files in binary mode by default
-  _fmode = _O_BINARY;
+  oncomplete_sym = NODE_PSYMBOL("oncomplete");
+
+#ifdef __POSIX__
+  StatWatcher::Initialize(target);
 #endif
 }
 
 }  // end namespace node
 
-NODE_MODULE(node_fs, node::InitFs);
+NODE_MODULE(node_fs, node::InitFs)
