@@ -27,19 +27,26 @@
 #include <mach/task.h>
 #include <mach/mach.h>
 #include <mach/mach_host.h>
-#include <mach-o/dyld.h> /* _NSGetExecutablePath */
 #include <limits.h> /* PATH_MAX */
 
 #include <unistd.h>  // sysconf
 #include <sys/param.h>
 #include <sys/sysctl.h>
 #include <time.h>
+#include <sys/socket.h>
+#include <net/if.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#include <ifaddrs.h>
+
+
 
 namespace node {
 
 using namespace v8;
 
 static char *process_title;
+double Platform::prog_start_time = Platform::GetUptime();
 
 char** Platform::SetupArgs(int argc, char *argv[]) {
   process_title = argc ? strdup(argv[0]) : NULL;
@@ -64,7 +71,7 @@ const char* Platform::GetProcessTitle(int *len) {
 
 // Researched by Tim Becker and Michael Knight
 // http://blog.kuriositaet.de/?p=257
-int Platform::GetMemory(size_t *rss, size_t *vsize) {
+int Platform::GetMemory(size_t *rss) {
   struct task_basic_info t_info;
   mach_msg_type_number_t t_info_count = TASK_BASIC_INFO_COUNT;
 
@@ -76,30 +83,11 @@ int Platform::GetMemory(size_t *rss, size_t *vsize) {
   if (r != KERN_SUCCESS) return -1;
 
   *rss = t_info.resident_size;
-  *vsize  = t_info.virtual_size;
 
   return 0;
 }
 
 
-int Platform::GetExecutablePath(char* buffer, size_t* size) {
-  uint32_t usize = *size;
-  int result = _NSGetExecutablePath(buffer, &usize);
-  if (result) return result;
-
-  char *path = new char[2*PATH_MAX];
-
-  char *fullpath = realpath(buffer, path);
-  if (fullpath == NULL) {
-    delete [] path;
-    return -1;
-  }
-  strncpy(buffer, fullpath, *size);
-  delete [] fullpath;
-  *size = strlen(buffer);
-  return 0;
-}
-
 int Platform::GetCPUInfo(Local<Array> *cpus) {
   Local<Object> cpuinfo;
   Local<Object> cputimes;
@@ -127,7 +115,7 @@ int Platform::GetCPUInfo(Local<Array> *cpus) {
     return -1;
   }
   *cpus = Array::New(numcpus);
-  for (int i = 0; i < numcpus; i++) {
+  for (unsigned int i = 0; i < numcpus; i++) {
     cpuinfo = Object::New();
     cputimes = Object::New();
     cputimes->Set(String::New("user"),
@@ -151,32 +139,7 @@ int Platform::GetCPUInfo(Local<Array> *cpus) {
   return 0;
 }
 
-double Platform::GetFreeMemory() {
-  double pagesize = static_cast<double>(sysconf(_SC_PAGESIZE));
-  vm_statistics_data_t info;
-  mach_msg_type_number_t count = sizeof(info) / sizeof(integer_t);
-
-  if (host_statistics(mach_host_self(), HOST_VM_INFO,
-                      (host_info_t)&info, &count) != KERN_SUCCESS) {
-    return -1;
-  }
-
-  return (static_cast<double>(info.free_count)) * pagesize;
-}
-
-double Platform::GetTotalMemory() {
-  uint64_t info;
-  static int which[] = {CTL_HW, HW_MEMSIZE};
-  size_t size = sizeof(info);
-
-  if (sysctl(which, 2, &info, &size, NULL, 0) < 0) {
-    return -1;
-  }
-
-  return static_cast<double>(info);
-}
-
-double Platform::GetUptime() {
+double Platform::GetUptimeImpl() {
   time_t now;
   struct timeval info;
   size_t size = sizeof(info);
@@ -190,22 +153,74 @@ double Platform::GetUptime() {
   return static_cast<double>(now - info.tv_sec);
 }
 
-int Platform::GetLoadAvg(Local<Array> *loads) {
-  struct loadavg info;
-  size_t size = sizeof(info);
-  static int which[] = {CTL_VM, VM_LOADAVG};
 
-  if (sysctl(which, 2, &info, &size, NULL, 0) < 0) {
-    return -1;
+v8::Handle<v8::Value> Platform::GetInterfaceAddresses() {
+  HandleScope scope;
+  struct ::ifaddrs *addrs, *ent;
+  struct ::sockaddr_in *in4;
+  struct ::sockaddr_in6 *in6;
+  char ip[INET6_ADDRSTRLEN];
+  Local<Object> ret, o;
+  Local<String> name, ipaddr, family;
+  Local<Array> ifarr;
+
+  if (getifaddrs(&addrs) != 0) {
+    return ThrowException(ErrnoException(errno, "getifaddrs"));
   }
-  (*loads)->Set(0, Number::New(static_cast<double>(info.ldavg[0])
-                               / static_cast<double>(info.fscale)));
-  (*loads)->Set(1, Number::New(static_cast<double>(info.ldavg[1])
-                               / static_cast<double>(info.fscale)));
-  (*loads)->Set(2, Number::New(static_cast<double>(info.ldavg[2])
-                               / static_cast<double>(info.fscale)));
 
-  return 0;
+  ret = Object::New();
+
+  for (ent = addrs; ent != NULL; ent = ent->ifa_next) {
+    bzero(&ip, sizeof (ip));
+    if (!(ent->ifa_flags & IFF_UP && ent->ifa_flags & IFF_RUNNING)) {
+      continue;
+    }
+
+    if (ent->ifa_addr == NULL) {
+      continue;
+    }
+
+    /*
+     * On Mac OS X getifaddrs returns information related to Mac Addresses for
+     * various devices, such as firewire, etc. These are not relevant here.
+     */
+    if (ent->ifa_addr->sa_family == AF_LINK)
+	    continue;
+
+    name = String::New(ent->ifa_name);
+    if (ret->Has(name)) {
+      ifarr = Local<Array>::Cast(ret->Get(name));
+    } else {
+      ifarr = Array::New();
+      ret->Set(name, ifarr);
+    }
+
+    if (ent->ifa_addr->sa_family == AF_INET6) {
+      in6 = (struct sockaddr_in6 *)ent->ifa_addr;
+      inet_ntop(AF_INET6, &(in6->sin6_addr), ip, INET6_ADDRSTRLEN);
+      family = String::New("IPv6");
+    } else if (ent->ifa_addr->sa_family == AF_INET) {
+      in4 = (struct sockaddr_in *)ent->ifa_addr;
+      inet_ntop(AF_INET, &(in4->sin_addr), ip, INET6_ADDRSTRLEN);
+      family = String::New("IPv4");
+    } else {
+      (void) strlcpy(ip, "<unknown sa family>", INET6_ADDRSTRLEN);
+      family = String::New("<unknown>");
+    }
+
+    o = Object::New();
+    o->Set(String::New("address"), String::New(ip));
+    o->Set(String::New("family"), family);
+    o->Set(String::New("internal"), ent->ifa_flags & IFF_LOOPBACK ?
+	True() : False());
+
+    ifarr->Set(ifarr->Length(), o);
+
+  }
+
+  freeifaddrs(addrs);
+
+  return scope.Close(ret);
 }
 
 }  // namespace node
